"""Data transformation utilities built on Apache Beam."""

from .beam_pipeline import BeamDataTransformer, BeamPipelineExecutor
from .config import (
    BigQuerySinkConfig,
    FileSinkConfig,
    FileSourceConfig,
    TransformerConfig,
)

__all__ = [
    "BeamDataTransformer",
    "BeamPipelineExecutor",
    "TransformerConfig",
    "FileSourceConfig",
    "FileSinkConfig",
    "BigQuerySinkConfig",
]



from __future__ import annotations

import argparse
import ast
import json
from dataclasses import dataclass
from functools import partial
from importlib import import_module
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Protocol, Sequence

from stscrapers.transformer.config import (
    BigQuerySinkConfig,
    FileSinkConfig,
    FileSourceConfig,
    TransformerConfig,
)


TransformFn = Callable[[Any], Any]


class PipelineExecutor(Protocol):
    def run(self, config: TransformerConfig, transform_fn: TransformFn) -> None:
        ...


@dataclass
class BeamPipelineExecutor:
    """Default pipeline executor that uses apache_beam under the hood."""

    pipeline_module: str = "apache_beam"

    def run(self, config: TransformerConfig, transform_fn: TransformFn) -> None:
        config.validate()
        beam = self._import_beam()
        options = self._build_pipeline_options(beam, config)
        with beam.Pipeline(options=options) as pipeline:
            input_pcoll = self._apply_read(beam, pipeline, config.source)
            transformed = transform_fn(input_pcoll)
            self._apply_writes(beam, transformed, config)

    def _import_beam(self):
        try:
            return import_module(self.pipeline_module)
        except ModuleNotFoundError as exc:  # pragma: no cover - depends on env
            raise ImportError(
                "apache_beam is not installed. Install it with "
                "`pip install apache-beam` to run transformations."
            ) from exc

    def _build_pipeline_options(self, beam, config: TransformerConfig):
        options = beam.options.pipeline_options.PipelineOptions()
        standard_options = options.view_as(
            beam.options.pipeline_options.StandardOptions
        )
        standard_options.runner = config.runner
        standard_options.streaming = config.streaming

        if config.runner == "DataflowRunner":
            gcloud_options = options.view_as(
                beam.options.pipeline_options.GoogleCloudOptions
            )
            gcloud_options.project = config.project
            gcloud_options.region = config.region
            gcloud_options.temp_location = config.temp_location
            if config.staging_location:
                gcloud_options.staging_location = config.staging_location

        setup_options = options.view_as(
            beam.options.pipeline_options.SetupOptions
        )
        setup_options.save_main_session = True
        return options

    def _apply_read(self, beam, pipeline, source: FileSourceConfig):
        #TODO: Add implementation for reading from Google Cloud Bucket
        compression = _resolve_compression_type(beam, source.compression)
        read_transform = beam.io.ReadFromText(
            source.resolved_path(),
            skip_header_lines=source.skip_header_lines,
            compression_type=compression,
        )
        return pipeline | "ReadInputFile" >> read_transform

    def _apply_writes(self, beam, pcoll, config: TransformerConfig) -> None:
        if config.output_file:
            self._write_file(beam, pcoll, config.output_file)
        if config.output_bigquery:
            self._write_bigquery(beam, pcoll, config)

    def _write_file(self, beam, pcoll, sink: FileSinkConfig) -> None:
        suffix = sink.file_name_suffix
        if suffix is None:
            suffix = f".{sink.file_format}"
        compression = _resolve_compression_type(beam, sink.compression)
        write_transform = beam.io.WriteToText(
            sink.resolved_path(),
            shard_name_template=sink.shard_name_template,
            file_name_suffix=suffix,
            compression_type=compression,
        )
        pcoll | "WriteOutputFile" >> write_transform

    def _write_bigquery(self, beam, pcoll, config: TransformerConfig) -> None:
        sink = config.output_bigquery
        assert sink is not None
        table_spec = sink.table_spec()
        write_transform = beam.io.WriteToBigQuery(
            table=table_spec,
            schema=sink.schema,
            create_disposition=sink.create_disposition,
            write_disposition=sink.write_disposition,
        )
        pcoll | "WriteBigQuery" >> write_transform


class BeamDataTransformer:
    """High-level façade that wires configs, transforms, and executors."""

    def __init__(
        self,
        config: TransformerConfig,
        transform_fn: TransformFn,
        executor: Optional[PipelineExecutor] = None,
    ) -> None:
        self.config = config
        self.transform_fn = transform_fn
        self.executor = executor or BeamPipelineExecutor()

    def process(self) -> None:
        """Execute the configured Beam pipeline."""
        self.executor.run(self.config, self.transform_fn)


def _resolve_compression_type(beam, value: Optional[str]):
    compression_types = beam.io.filesystem.CompressionTypes
    if not value:
        return compression_types.AUTO
    if isinstance(value, str):
        normalized = value.upper()
        if hasattr(compression_types, normalized):
            return getattr(compression_types, normalized)
        raise ValueError(f"Unsupported compression type: {value}")
    return value


def parse_param_overrides(entries: Sequence[str] | None) -> Dict[str, Any]:
    params: Dict[str, Any] = {}
    if not entries:
        return params
    for entry in entries:
        if "=" not in entry:
            raise ValueError(f"Invalid parameter '{entry}'. Use key=value format.")
        key, value = entry.split("=", 1)
        params[key] = _coerce_cli_value(value)
    return params


def load_transform_callable(path: str, params: Dict[str, Any]) -> TransformFn:
    if ":" not in path:
        raise ValueError(
            "Transform path must use module:function format "
            f"(got '{path}')"
        )
    module_name, func_name = path.split(":", 1)
    module = import_module(module_name)
    transform_fn = getattr(module, func_name)
    if params:
        transform_fn = partial(transform_fn, **params)
    return transform_fn


def build_config_from_dict(raw: Dict[str, Any]) -> TransformerConfig:
    source_dict = raw.get("source") or {}
    if "path" not in source_dict:
        raise ValueError("Configuration file must include source.path")
    source = FileSourceConfig(**_filter_kwargs(source_dict, FileSourceConfig))

    output_file_dict = raw.get("output_file")
    output_file = (
        FileSinkConfig(**_filter_kwargs(output_file_dict, FileSinkConfig))
        if output_file_dict
        else None
    )

    output_bq_dict = raw.get("output_bigquery")
    output_bigquery = (
        BigQuerySinkConfig(**_filter_kwargs(output_bq_dict, BigQuerySinkConfig))
        if output_bq_dict
        else None
    )

    return TransformerConfig(
        source=source,
        runner=raw.get("runner", "DirectRunner"),
        output_file=output_file,
        output_bigquery=output_bigquery,
        project=raw.get("project"),
        region=raw.get("region"),
        temp_location=raw.get("temp_location"),
        staging_location=raw.get("staging_location"),
        streaming=raw.get("streaming", False),
    )


def build_config_from_args(
    args: argparse.Namespace,
) -> tuple[TransformerConfig, Optional[Dict[str, Any]]]:
    config: Optional[TransformerConfig] = None
    config_data: Optional[Dict[str, Any]] = None
    if args.config:
        config_data = _load_config_file(args.config)
        config = build_config_from_dict(config_data)
    if config is None:
        config = _build_config_from_cli(args)
    _apply_cli_overrides(config, args)
    return config, config_data


def _build_config_from_cli(args: argparse.Namespace) -> TransformerConfig:
    if not args.source_path:
        raise ValueError(
            "A source path is required when no configuration file is provided."
        )
    source = FileSourceConfig(
        path=args.source_path,
        file_format=args.source_format or "csv",
        skip_header_lines=args.source_skip_header_lines or 0,
        compression=args.source_compression,
        is_remote=args.source_remote or False,
    )

    output_file = None
    if args.output_file_path:
        output_file = FileSinkConfig(
            path=args.output_file_path,
            file_format=args.output_file_format or "csv",
            shard_name_template=args.output_file_shard_template or "",
            file_name_suffix=args.output_file_suffix,
            compression=args.output_file_compression,
            is_remote=args.output_file_remote or False,
        )

    output_bigquery = None
    if args.output_bigquery_table:
        dataset = args.output_bigquery_dataset
        schema = args.output_bigquery_schema
        if not (dataset and schema):
            raise ValueError(
                "BigQuery output requires --output-bigquery-dataset and "
                "--output-bigquery-schema."
            )
        output_bigquery = BigQuerySinkConfig(
            dataset=dataset,
            table=args.output_bigquery_table,
            schema=schema,
            project=args.output_bigquery_project or args.project,
        )

    if not (output_file or output_bigquery):
        raise ValueError(
            "Specify at least one output via --output-file-path or "
            "--output-bigquery-table."
        )

    streaming = args.streaming if args.streaming is not None else False
    return TransformerConfig(
        source=source,
        runner=args.runner or "DirectRunner",
        output_file=output_file,
        output_bigquery=output_bigquery,
        project=args.project,
        region=args.region,
        temp_location=args.temp_location,
        staging_location=args.staging_location,
        streaming=streaming,
    )


def _apply_cli_overrides(config: TransformerConfig, args: argparse.Namespace) -> None:
    if args.source_path:
        config.source.path = args.source_path
    if args.source_format:
        config.source.file_format = args.source_format
    if args.source_skip_header_lines is not None:
        config.source.skip_header_lines = args.source_skip_header_lines
    if args.source_compression:
        config.source.compression = args.source_compression
    if args.source_remote is not None:
        config.source.is_remote = args.source_remote

    if args.runner:
        config.runner = args.runner
    if args.project:
        config.project = args.project
    if args.region:
        config.region = args.region
    if args.temp_location:
        config.temp_location = args.temp_location
    if args.staging_location:
        config.staging_location = args.staging_location
    if args.streaming is not None:
        config.streaming = args.streaming

    _override_output_file(config, args)
    _override_bigquery(config, args)


def _override_output_file(config: TransformerConfig, args: argparse.Namespace) -> None:
    overrides_present = any(
        [
            args.output_file_path,
            args.output_file_format,
            args.output_file_shard_template,
            args.output_file_suffix,
            args.output_file_compression,
            args.output_file_remote is not None,
        ]
    )
    if not overrides_present:
        return

    if config.output_file is None:
        if not args.output_file_path:
            raise ValueError(
                "Creating a file output via CLI overrides requires "
                "--output-file-path."
            )
        config.output_file = FileSinkConfig(path=args.output_file_path)

    if args.output_file_path:
        config.output_file.path = args.output_file_path
    if args.output_file_format:
        config.output_file.file_format = args.output_file_format
    if args.output_file_shard_template is not None:
        config.output_file.shard_name_template = args.output_file_shard_template
    if args.output_file_suffix is not None:
        config.output_file.file_name_suffix = args.output_file_suffix
    if args.output_file_compression is not None:
        config.output_file.compression = args.output_file_compression
    if args.output_file_remote is not None:
        config.output_file.is_remote = args.output_file_remote


def _override_bigquery(config: TransformerConfig, args: argparse.Namespace) -> None:
    overrides_present = any(
        [
            args.output_bigquery_dataset,
            args.output_bigquery_table,
            args.output_bigquery_schema,
            args.output_bigquery_project,
        ]
    )
    if not overrides_present:
        return

    if config.output_bigquery is None:
        required = {
            "dataset": args.output_bigquery_dataset,
            "table": args.output_bigquery_table,
            "schema": args.output_bigquery_schema,
        }
        missing = [key for key, value in required.items() if not value]
        if missing:
            raise ValueError(
                "Creating a BigQuery sink via overrides requires: "
                + ", ".join(missing)
            )
        config.output_bigquery = BigQuerySinkConfig(
            dataset=required["dataset"],
            table=required["table"],
            schema=required["schema"],
            project=args.output_bigquery_project or args.project or config.project,
        )
    else:
        if args.output_bigquery_dataset:
            config.output_bigquery.dataset = args.output_bigquery_dataset
        if args.output_bigquery_table:
            config.output_bigquery.table = args.output_bigquery_table
        if args.output_bigquery_schema:
            config.output_bigquery.schema = args.output_bigquery_schema
        if args.output_bigquery_project:
            config.output_bigquery.project = args.output_bigquery_project


def _filter_kwargs(data: Dict[str, Any], cls: type) -> Dict[str, Any]:
    allowed = set(cls.__dataclass_fields__.keys())  # type: ignore[attr-defined]
    return {key: value for key, value in (data or {}).items() if key in allowed}


def _coerce_cli_value(value: str) -> Any:
    lowered = value.lower()
    if lowered == "true":
        return True
    if lowered == "false":
        return False
    if lowered == "null":
        return None
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError):
        return value


def _load_config_file(path: str) -> Dict[str, Any]:
    file_path = Path(path)
    text = file_path.read_text(encoding="utf-8")
    suffix = file_path.suffix.lower()
    if suffix in {".yaml", ".yml"}:
        try:
            import yaml  # type: ignore
        except ImportError as exc:  # pragma: no cover - optional dependency
            raise ImportError(
                "PyYAML is required to load YAML configs. Install with `pip install pyyaml`."
            ) from exc
        data = yaml.safe_load(text)
        return data or {}
    return json.loads(text)


def _build_cli_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Run an Apache Beam transformation using the configured transformer."
    )
    parser.add_argument(
        "--config",
        help="Path to a JSON or YAML config file describing the transformer.",
    )
    parser.add_argument(
        "--transform",
        help="Python path to the transform callable (e.g. module.sub:build_transform).",
    )
    parser.add_argument(
        "--param",
        action="append",
        default=[],
        help="Parameter passed to the transform callable (key=value; repeatable).",
    )
    parser.add_argument("--runner", help="Beam runner (DirectRunner, DataflowRunner, etc.)")
    parser.add_argument("--project", help="GCP project (required for Dataflow).")
    parser.add_argument("--region", help="GCP region (Dataflow).")
    parser.add_argument("--temp-location", help="Temp location for Dataflow staging.")
    parser.add_argument("--staging-location", help="Optional Dataflow staging location.")

    parser.add_argument("--source-path", help="Input file path or URI.")
    parser.add_argument("--source-format", help="Input file format (default csv).")
    parser.add_argument(
        "--source-skip-header-lines",
        type=int,
        help="Number of header lines to skip when reading text files.",
    )
    parser.add_argument("--source-compression", help="Compression codec for the input file.")
    parser.add_argument(
        "--source-remote",
        dest="source_remote",
        action="store_true",
        help="Treat the source path as remote (e.g., gs://).",
    )
    parser.add_argument(
        "--source-local",
        dest="source_remote",
        action="store_false",
        help="Treat the source path as local.",
    )

    parser.add_argument("--output-file-path", help="Destination file path or prefix.")
    parser.add_argument("--output-file-format", help="Output file format (default csv).")
    parser.add_argument(
        "--output-file-shard-template",
        help="Custom shard template for Beam WriteToText.",
    )
    parser.add_argument("--output-file-suffix", help="File suffix (default uses format).")
    parser.add_argument("--output-file-compression", help="Compression codec for output file.")
    parser.add_argument(
        "--output-file-remote",
        dest="output_file_remote",
        action="store_true",
        help="Treat the output file path as remote.",
    )
    parser.add_argument(
        "--output-file-local",
        dest="output_file_remote",
        action="store_false",
        help="Treat the output file path as local.",
    )

    parser.add_argument("--output-bigquery-dataset", help="BigQuery dataset name.")
    parser.add_argument("--output-bigquery-table", help="BigQuery table name.")
    parser.add_argument("--output-bigquery-schema", help="BigQuery schema definition.")
    parser.add_argument("--output-bigquery-project", help="Overrides the BigQuery project.")

    parser.add_argument(
        "--streaming",
        dest="streaming",
        action="store_true",
        help="Run the pipeline in streaming mode.",
    )
    parser.add_argument(
        "--batch",
        dest="streaming",
        action="store_false",
        help="Force batch mode when overriding a config file.",
    )

    parser.set_defaults(
        source_remote=None,
        output_file_remote=None,
        streaming=None,
    )

    return parser


def main(argv: Optional[Sequence[str]] = None) -> int:
    parser = _build_cli_parser()
    args = parser.parse_args(argv)
    config, config_data = build_config_from_args(args)
    params = parse_param_overrides(args.param)
    transform_path = args.transform or (config_data or {}).get("transform")
    if not transform_path:
        parser.error("A transform callable must be provided via --transform or config file.")
    transform_fn = load_transform_callable(transform_path, params)
    transformer = BeamDataTransformer(config=config, transform_fn=transform_fn)
    transformer.process()
    return 0


if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())



"""Storage backend implementations for scrapers and transformers."""

from .backends import (
    GcsStorageBackend,
    LocalFileStorageBackend,
    StorageBackend,
)

__all__ = ["LocalFileStorageBackend", "GcsStorageBackend", "StorageBackend"]



"""Common scraper interfaces and implementations."""

from .base import DatasetScraper
from .models import (
    DatasetCheckResult,
    DatasetConfig,
    DatasetFetchResult,
    DatasetUpdateResult,
    DownloadedResource,
)
from .socrata import SocrataDatasetScraper

__all__ = [
    "DatasetScraper",
    "DatasetCheckResult",
    "DatasetConfig",
    "DatasetFetchResult",
    "DatasetUpdateResult",
    "DownloadedResource",
    "SocrataDatasetScraper",
]



from __future__ import annotations

import json
import logging
from datetime import datetime
from typing import Any, Dict, Optional, Sequence
from urllib import error as urllib_error, parse as urllib_parse, request as urllib_request

from stscrapers.scrapers.base import DatasetScraper
from stscrapers.scrapers.models import (
    DatasetCheckResult,
    DatasetConfig,
    DatasetFetchResult,
    DownloadedResource,
)


logger = logging.getLogger(__name__)


class HttpClient:
    """Minimal HTTP helper to allow dependency injection in tests."""

    def __init__(self, timeout: int = 30) -> None:
        self.timeout = timeout

    def get_json(self, url: str, headers: Optional[Dict[str, str]] = None) -> Any:
        payload = self._execute(url, headers)
        return json.loads(payload.decode("utf-8"))

    def get_bytes(self, url: str, headers: Optional[Dict[str, str]] = None) -> bytes:
        return self._execute(url, headers)

    def _execute(self, url: str, headers: Optional[Dict[str, str]]) -> bytes:
        req = urllib_request.Request(url, headers=headers or {})
        try:
            with urllib_request.urlopen(req, timeout=self.timeout) as resp:
                return resp.read()
        except urllib_error.HTTPError as exc:  # pragma: no cover - network dependent
            raise RuntimeError(f"HTTP error {exc.code} for {url}") from exc
        except urllib_error.URLError as exc:  # pragma: no cover - network dependent
            raise RuntimeError(f"Failed to reach {url}: {exc.reason}") from exc


class SocrataDatasetScraper(DatasetScraper):
    """
    Socrata implementation of the DatasetScraper base interface.
    Documentation for Socrata API: 
    - https://dev.socrata.com/docs/other/discovery#?route=get-/catalog/v1-ids-4x4-
    - https://dev.socrata.com/docs/endpoints.html
    """

    name = "socrata"
    SUPPORTED_FORMATS = {"csv", "json", "tsv", "geojson"}

    def __init__(
        self,
        discovery_base_url: str = "https://api.us.socrata.com/api/catalog/v1",
        http_client: Optional[HttpClient] = None,
        app_token: Optional[str] = None,
    ) -> None:
        self.discovery_base_url = discovery_base_url.rstrip("/")
        self.http_client = http_client or HttpClient()
        self.app_token = app_token

    def check_updates(self, dataset: DatasetConfig) -> DatasetCheckResult:
        dataset_domain = _normalize_domain(dataset.domain)
        params = urllib_parse.urlencode(
            {"ids": f"{dataset_domain}/{dataset.identifier}", "only": "datasets"}
        )
        url = f"{self.discovery_base_url}?{params}"
        payload = self.http_client.get_json(url, headers=self._build_headers())
        results = payload.get("results") or []
        if not results:
            raise ValueError(f"Dataset {dataset.identifier} not found via discovery API")

        resource = results[0].get("resource", {})
        remote_ts = (
            resource.get("dataUpdatedAt")
            or resource.get("updatedAt")
            or resource.get("metadataUpdatedAt")
        )
        remote_dt = _parse_datetime(remote_ts)
        has_updates = False
        if remote_dt and dataset.last_local_update:
            has_updates = remote_dt > dataset.last_local_update
        elif remote_dt and not dataset.last_local_update:
            has_updates = True

        metadata = {
            "name": resource.get("name"),
            "permalink": results[0].get("permalink"),
            "link": results[0].get("link"),
            "classification": results[0].get("classification"),
            "resource": resource,
        }
        return DatasetCheckResult(
            dataset_id=dataset.identifier,
            has_updates=has_updates,
            remote_updated_at=remote_dt,
            metadata=metadata,
        )

    def fetch_dataset(self, dataset: DatasetConfig) -> DatasetFetchResult:
        formats = list(dataset.download_formats or ["csv"])
        files = []
        for fmt in formats:
            normalized = fmt.lower()
            if normalized not in self.SUPPORTED_FORMATS:
                raise ValueError(
                    f"Format '{fmt}' is not supported. "
                    f"Supported: {sorted(self.SUPPORTED_FORMATS)}"
                )
            url = self._build_export_url(dataset.domain, dataset.identifier, normalized)
            content = self.http_client.get_bytes(url, headers=self._build_headers())
            files.append(
                DownloadedResource(
                    format=normalized,
                    content=content,
                    source_url=url,
                )
            )

        metadata = {
            "dataset_id": dataset.identifier,
            "domain": dataset.domain,
            "download_formats": formats,
        }
        return DatasetFetchResult(
            dataset_id=dataset.identifier,
            files=files,
            metadata=metadata,
        )

    def _build_headers(self) -> Dict[str, str]:
        headers: Dict[str, str] = {}
        if self.app_token:
            headers["X-App-Token"] = self.app_token
        return headers

    @staticmethod
    def _build_export_url(domain: str, dataset_id: str, fmt: str) -> str:
        normalized_domain = _normalize_domain(domain)
        return (
            f"https://{normalized_domain}/api/views/{dataset_id}/rows."
            f"{fmt}?accessType=DOWNLOAD"
        )


def _parse_datetime(value: Optional[str]) -> Optional[datetime]:
    if not value:
        return None
    try:
        sanitized = value.replace("Z", "+00:00")
        return datetime.fromisoformat(sanitized)
    except ValueError:
        logger.debug("Unable to parse datetime value from %s", value)
        return None


def _normalize_domain(domain: str) -> str:
    stripped = domain.strip()
    if not stripped:
        raise ValueError("Domain cannot be empty")
    parsed = urllib_parse.urlparse(stripped if "://" in stripped else f"//{stripped}")
    host = parsed.netloc or parsed.path
    return host.rstrip("/")



from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, Optional, Protocol

from stscrapers.scrapers.models import DatasetConfig


class StorageBackend(Protocol):
    def write_dataset_file(
        self, dataset: DatasetConfig, file_format: str, payload: bytes
    ) -> Path:
        ...

    def write_metadata(self, dataset: DatasetConfig, metadata: Dict[str, object]) -> Path:
        ...


class LocalFileStorageBackend:
    """Local filesystem storage backend (default)."""

    def __init__(self, base_directory: Optional[Path] = None) -> None:
        self.base_directory = Path(base_directory) if base_directory else None

    def _resolve(self, target: Path) -> Path:
        return (self.base_directory / target) if self.base_directory else target

    def _ensure_dir(self, target: Path) -> None:
        target.mkdir(parents=True, exist_ok=True)

    def write_dataset_file(
        self, dataset: DatasetConfig, file_format: str, payload: bytes
    ) -> Path:
        directory = self._resolve(dataset.local_path)
        self._ensure_dir(directory)
        file_path = directory / f"{dataset.identifier}.{file_format}"
        file_path.write_bytes(payload)
        return file_path

    def write_metadata(self, dataset: DatasetConfig, metadata: Dict[str, object]) -> Path:
        directory = self._resolve(dataset.local_path)
        self._ensure_dir(directory)
        filename = dataset.metadata_filename or f"{dataset.identifier}.metadata.json"
        path = directory / filename
        path.write_text(json.dumps(metadata, indent=2), encoding="utf-8")
        return path


class GcsStorageBackend:
    """Google Cloud Storage backend for storing dataset files."""

    def __init__(
        self,
        bucket_name: str,
        client: Optional[object] = None,
        prefix: Optional[str] = None,
    ) -> None:
        self.bucket_name = bucket_name
        self.prefix = (prefix or "").strip("/")
        if client is None:
            try:
                from google.cloud import storage
            except ImportError as exc:  # pragma: no cover - import guard
                raise ImportError(
                    "google-cloud-storage is required for GcsStorageBackend. "
                    "Install with `pip install google-cloud-storage`."
                ) from exc
            client = storage.Client()
        self._client = client
        self._bucket = self._client.bucket(bucket_name)

    def write_dataset_file(
        self, dataset: DatasetConfig, file_format: str, payload: bytes
    ) -> Path:
        filename = f"{dataset.identifier}.{file_format}"
        blob_path = self._blob_path(dataset, filename)
        self._upload(blob_path, payload, content_type=self._guess_content_type(file_format))
        return Path(f"gs://{self.bucket_name}/{blob_path}")

    def write_metadata(self, dataset: DatasetConfig, metadata: Dict[str, object]) -> Path:
        filename = dataset.metadata_filename or f"{dataset.identifier}.metadata.json"
        blob_path = self._blob_path(dataset, filename)
        payload = json.dumps(metadata, indent=2).encode("utf-8")
        self._upload(blob_path, payload, content_type="application/json")
        return Path(f"gs://{self.bucket_name}/{blob_path}")

    def _upload(self, blob_name: str, payload: bytes, content_type: Optional[str]) -> None:
        blob = self._bucket.blob(blob_name)
        blob.upload_from_string(payload, content_type=content_type)

    def _blob_path(self, dataset: DatasetConfig, filename: str) -> str:
        parts = [
            self.prefix,
            str(dataset.local_path).strip("/"),
            filename,
        ]
        return "/".join(part for part in parts if part)

    @staticmethod
    def _guess_content_type(file_format: str) -> Optional[str]:
        mapping = {
            "csv": "text/csv",
            "tsv": "text/tab-separated-values",
            "json": "application/json",
            "geojson": "application/geo+json",
        }
        return mapping.get(file_format.lower())



from __future__ import annotations

import csv
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, Iterable, Optional, Protocol

from stscrapers.scrapers.models import DatasetCheckResult, DatasetConfig, DatasetUpdateResult


class MetadataLogger(Protocol):
    def record_check(
        self, dataset: DatasetConfig, result: DatasetCheckResult
    ) -> None:
        ...

    def record_update(
        self, dataset: DatasetConfig, result: DatasetUpdateResult
    ) -> None:
        ...


class LocalCsvMetadataLogger:
    """CSV-backed metadata log with rows per check/update event."""

    HEADERS = [
        "timestamp",
        "dataset_id",
        "action",
        "has_updates",
        "remote_updated_at",
        "file_format",
        "file_path",
        "file_bytes",
        "source_url",
    ]

    def __init__(self, log_filename: str = "metadata_log.csv") -> None:
        self.log_filename = log_filename

    def record_check(
        self, dataset: DatasetConfig, result: DatasetCheckResult
    ) -> None:
        row = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "dataset_id": dataset.identifier,
            "action": "check",
            "has_updates": str(result.has_updates),
            "remote_updated_at": (
                result.remote_updated_at.isoformat()
                if result.remote_updated_at
                else ""
            ),
            "file_format": "",
            "file_path": "",
            "file_bytes": "",
            "source_url": "",
        }
        self._write_rows(dataset, [row])

    def record_update(
        self, dataset: DatasetConfig, result: DatasetUpdateResult
    ) -> None:
        timestamp = datetime.now(timezone.utc).isoformat()
        rows = []
        for file_info in result.files:
            rows.append(
                {
                    "timestamp": timestamp,
                    "dataset_id": dataset.identifier,
                    "action": "update",
                    "has_updates": "",
                    "remote_updated_at": "",
                    "file_format": file_info.get("format", ""),
                    "file_path": file_info.get("path", ""),
                    "file_bytes": str(file_info.get("bytes", "")),
                    "source_url": file_info.get("source_url", ""),
                }
            )
        self._write_rows(dataset, rows)

    def _write_rows(
        self, dataset: DatasetConfig, rows: Iterable[Dict[str, str]]
    ) -> None:
        dataset.local_path.mkdir(parents=True, exist_ok=True)
        log_path = dataset.local_path / self.log_filename
        exists = log_path.exists()
        with log_path.open("a", encoding="utf-8", newline="") as handle:
            writer = csv.DictWriter(handle, fieldnames=self.HEADERS)
            if not exists:
                writer.writeheader()
            for row in rows:
                writer.writerow(row)


class CloudMetadataLogger(MetadataLogger):
    """Placeholder for future cloud metadata logging implementation."""

    def record_check(
        self, dataset: DatasetConfig, result: DatasetCheckResult
    ) -> None:  # pragma: no cover - placeholder
        raise NotImplementedError("Cloud metadata logging is not implemented yet.")

    def record_update(
        self, dataset: DatasetConfig, result: DatasetUpdateResult
    ) -> None:  # pragma: no cover - placeholder
        raise NotImplementedError("Cloud metadata logging is not implemented yet.")


class DatasetStatusLogger(Protocol):
    def record_check(
        self, dataset: DatasetConfig, result: DatasetCheckResult
    ) -> None:
        ...

    def record_update(
        self, dataset: DatasetConfig, result: DatasetUpdateResult
    ) -> None:
        ...


class LocalDatasetStatusLogger(DatasetStatusLogger):
    """Tracks per-dataset status in a CSV for quick sync monitoring."""

    HEADERS = [
        "dataset_id",
        "domain",
        "dataset_name",
        "dataset_link",
        "local_path",
        "last_checked_at",
        "last_updated_at",
        "remote_updated_at",
    ]

    def __init__(
        self,
        log_directory: Optional[Path] = None,
        log_filename: str = "dataset_status.csv",
    ) -> None:
        self.log_directory = Path(log_directory) if log_directory else None
        self.log_filename = log_filename

    def record_check(
        self, dataset: DatasetConfig, result: DatasetCheckResult
    ) -> None:
        log_path = self._resolve_path(dataset)
        rows = self._load_rows(log_path)
        row = self._get_or_create_row(rows, dataset)
        now = datetime.now(timezone.utc).isoformat()
        row["domain"] = _normalize_domain(dataset.domain)
        row["dataset_name"] = result.metadata.get("name") or result.metadata.get(
            "resource", {}
        ).get("name", "")
        row["dataset_link"] = (
            result.metadata.get("link")
            or row["dataset_link"]
            or self._default_link(dataset)
        )
        row["local_path"] = str(dataset.local_path)
        row["last_checked_at"] = now
        row["remote_updated_at"] = (
            result.remote_updated_at.isoformat()
            if result.remote_updated_at
            else row.get("remote_updated_at", "")
        )
        self._write_rows(log_path, rows)

    def record_update(
        self, dataset: DatasetConfig, result: DatasetUpdateResult
    ) -> None:
        log_path = self._resolve_path(dataset)
        rows = self._load_rows(log_path)
        row = self._get_or_create_row(rows, dataset)
        now = datetime.now(timezone.utc).isoformat()
        row["domain"] = _normalize_domain(dataset.domain)
        row["dataset_link"] = row.get("dataset_link") or self._default_link(dataset)
        row["local_path"] = str(dataset.local_path)
        row["last_updated_at"] = now
        self._write_rows(log_path, rows)

    def _resolve_path(self, dataset: DatasetConfig) -> Path:
        base_dir = self.log_directory or dataset.local_path
        base_dir.mkdir(parents=True, exist_ok=True)
        return base_dir / self.log_filename

    def _load_rows(self, path: Path) -> list[Dict[str, str]]:
        if not path.exists():
            return []
        with path.open("r", encoding="utf-8", newline="") as handle:
            reader = csv.DictReader(handle)
            return [dict(row) for row in reader]

    def _write_rows(self, path: Path, rows: Iterable[Dict[str, str]]) -> None:
        with path.open("w", encoding="utf-8", newline="") as handle:
            writer = csv.DictWriter(handle, fieldnames=self.HEADERS)
            writer.writeheader()
            for row in rows:
                writer.writerow(row)

    def _get_or_create_row(
        self, rows: list[Dict[str, str]], dataset: DatasetConfig
    ) -> Dict[str, str]:
        for row in rows:
            if row.get("dataset_id") == dataset.identifier:
                return row
        row = {header: "" for header in self.HEADERS}
        row["dataset_id"] = dataset.identifier
        rows.append(row)
        return row

    @staticmethod
    def _default_link(dataset: DatasetConfig) -> str:
        domain = _normalize_domain(dataset.domain)
        return f"https://{domain}/d/{dataset.identifier}"


class CloudDatasetStatusLogger(DatasetStatusLogger):
    """Placeholder for a future cloud dataset status logger."""

    def record_check(
        self, dataset: DatasetConfig, result: DatasetCheckResult
    ) -> None:  # pragma: no cover - placeholder
        raise NotImplementedError(
            "Cloud dataset status logging is not implemented yet."
        )

    def record_update(
        self, dataset: DatasetConfig, result: DatasetUpdateResult
    ) -> None:  # pragma: no cover - placeholder
        raise NotImplementedError(
            "Cloud dataset status logging is not implemented yet."
        )


def _normalize_domain(domain: str) -> str:
    stripped = domain.strip()
    if not stripped:
        raise ValueError("Domain cannot be empty")
    if "://" not in stripped:
        stripped = f"//{stripped}"
    from urllib import parse

    parsed = parse.urlparse(stripped)
    host = parsed.netloc or parsed.path
    return host.rstrip("/")



"""Toy Beam transforms for demonstration and testing."""

from __future__ import annotations

try:  # pragma: no cover - optional dependency import
    import apache_beam as beam
except ImportError as exc:  # pragma: no cover - import guard
    raise ImportError(
        "apache_beam is required to use the sample transforms. "
        "Install it with `pip install apache-beam`."
    ) from exc


def clean_csv(pcoll, suffix: str = ""):
    """Normalize string rows and emit dictionaries with the text/length metadata."""

    return (
        pcoll
        | "StripWhitespace" >> beam.Map(lambda line: line.strip())
        | "FilterEmpty" >> beam.Filter(bool)
        | "Uppercase" >> beam.Map(str.upper)
        | "AppendSuffix" >> beam.Map(lambda line, suf: f"{line}{suf}", suffix)
        | "ToDict" >> beam.Map(lambda line: {"text": line, "length": len(line)})
    )


__all__ = ["clean_csv"]



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Sequence


@dataclass
class DatasetConfig:
    identifier: str
    domain: str
    local_path: Path
    last_local_update: Optional[datetime] = None
    download_formats: Optional[Sequence[str]] = None
    metadata_filename: Optional[str] = None
    extra: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DatasetCheckResult:
    dataset_id: str
    has_updates: bool
    remote_updated_at: Optional[datetime]
    metadata: Dict[str, Any]


@dataclass
class DownloadedResource:
    format: str
    content: bytes
    source_url: str


@dataclass
class DatasetFetchResult:
    dataset_id: str
    files: Sequence[DownloadedResource]
    metadata: Dict[str, Any]


@dataclass
class DatasetUpdateResult:
    dataset_id: str
    files: Sequence[Dict[str, Any]]
    metadata_path: Path



from __future__ import annotations

from typing import Protocol

from stscrapers.scrapers.models import (
    DatasetCheckResult,
    DatasetConfig,
    DatasetFetchResult,
)


class DatasetScraper(Protocol):
    """Base interface every scraper must implement."""

    name: str

    def check_updates(self, dataset: DatasetConfig) -> DatasetCheckResult:
        """Return metadata describing whether the dataset has changed."""

    def fetch_dataset(self, dataset: DatasetConfig) -> DatasetFetchResult:
        """Download dataset payloads but do not persist them."""



import csv
import json
import tempfile
import unittest
from datetime import datetime, timezone
from pathlib import Path

from stscrapers.scrapers.manager import ScraperManager
from stscrapers.scrapers.models import DatasetConfig
from stscrapers.scrapers.socrata import SocrataDatasetScraper


class FakeHttpClient:
    def __init__(self) -> None:
        self.json_payloads = {}
        self.byte_payloads = {}

    def add_json(self, url: str, payload) -> None:
        self.json_payloads[url] = payload

    def add_bytes(self, url: str, payload: bytes) -> None:
        self.byte_payloads[url] = payload

    def get_json(self, url, headers=None):
        return self.json_payloads[url]

    def get_bytes(self, url, headers=None):
        return self.byte_payloads[url]


class ScraperManagerTests(unittest.TestCase):
    def setUp(self) -> None:
        self.manager = ScraperManager()
        self.fake_client = FakeHttpClient()

    def test_check_dataset_records_logs(self) -> None:
        monitor_url = (
            "https://api.us.socrata.com/api/catalog/v1?"
            "ids=data.example.com%2Fabcd-1234&only=datasets"
        )
        self.fake_client.add_json(
            monitor_url,
            {
                "results": [
                    {
                        "permalink": "https://data.example.com/d/abcd-1234",
                        "link": "https://data.example.com/d/abcd-1234",
                        "classification": {"domain_category": "test"},
                        "resource": {
                            "id": "abcd-1234",
                            "name": "Example Dataset",
                            "dataUpdatedAt": "2022-01-01T00:00:00Z",
                        },
                    }
                ]
            },
        )
        scraper = SocrataDatasetScraper(http_client=self.fake_client)
        with tempfile.TemporaryDirectory() as tmpdir:
            dataset = DatasetConfig(
                identifier="abcd-1234",
                domain="data.example.com",
                local_path=Path(tmpdir),
                last_local_update=datetime(2020, 1, 1, tzinfo=timezone.utc),
            )
            result = self.manager.check_dataset(scraper, dataset)

            self.assertTrue(result.has_updates)
            metadata_log = dataset.local_path / "metadata_log.csv"
            self.assertTrue(metadata_log.exists())
            with metadata_log.open() as handle:
                rows = list(csv.DictReader(handle))
            self.assertEqual(rows[-1]["action"], "check")
            status_log = dataset.local_path / "dataset_status.csv"
            with status_log.open() as handle:
                status_rows = list(csv.DictReader(handle))
            self.assertEqual(status_rows[-1]["dataset_id"], "abcd-1234")
            self.assertTrue(status_rows[-1]["last_checked_at"])

    def test_update_dataset_persists_files_and_updates_logs(self) -> None:
        download_url = (
            "https://data.example.com/api/views/abcd-1234/rows.csv?accessType=DOWNLOAD"
        )
        payload = b"col1,col2\n1,2\n"
        self.fake_client.add_bytes(download_url, payload)
        scraper = SocrataDatasetScraper(http_client=self.fake_client)
        with tempfile.TemporaryDirectory() as tmpdir:
            dataset = DatasetConfig(
                identifier="abcd-1234",
                domain="data.example.com",
                local_path=Path(tmpdir),
                download_formats=["csv"],
            )
            result = self.manager.update_dataset(scraper, dataset)

            self.assertEqual(len(result.files), 1)
            stored_file = Path(result.files[0]["path"])
            self.assertTrue(stored_file.exists())
            self.assertEqual(stored_file.read_bytes(), payload)
            metadata_path = dataset.local_path / "abcd-1234.metadata.json"
            self.assertTrue(metadata_path.exists())
            metadata = json.loads(metadata_path.read_text())
            self.assertEqual(metadata["files"][0]["bytes"], len(payload))
            metadata_log = dataset.local_path / "metadata_log.csv"
            with metadata_log.open() as handle:
                rows = list(csv.DictReader(handle))
            self.assertTrue(any(row["action"] == "update" for row in rows))
            status_log = dataset.local_path / "dataset_status.csv"
            with status_log.open() as handle:
                status_rows = list(csv.DictReader(handle))
            self.assertTrue(status_rows[-1]["last_updated_at"])


if __name__ == "__main__":
    unittest.main()



import json
import os
import tempfile
import unittest
from types import SimpleNamespace
from unittest import mock

from stscrapers.transformer import (
    BeamDataTransformer,
    BeamPipelineExecutor,
    FileSinkConfig,
    FileSourceConfig,
    TransformerConfig,
)
from stscrapers.transformer.beam_pipeline import (
    build_config_from_args,
    load_transform_callable,
    parse_param_overrides,
)


class DummyExecutor:
    def __init__(self) -> None:
        self.calls = 0

    def run(self, config, transform_fn):
        self.calls += 1
        self.config = config
        self.transform_fn = transform_fn


class TransformerTests(unittest.TestCase):
    def _config(self) -> TransformerConfig:
        return TransformerConfig(
            source=FileSourceConfig(path="input.csv"),
            output_file=FileSinkConfig(path="output.csv"),
        )

    def test_transformer_delegates_to_executor(self) -> None:
        executor = DummyExecutor()
        config = self._config()

        transformer = BeamDataTransformer(
            config=config, transform_fn=lambda pcoll: pcoll, executor=executor
        )
        transformer.process()

        self.assertEqual(executor.calls, 1)
        self.assertIs(executor.config, config)

    def test_dataflow_runner_requires_project_and_locations(self) -> None:
        config = TransformerConfig(
            source=FileSourceConfig(path="input.csv"),
            output_file=FileSinkConfig(path="output.csv"),
            runner="DataflowRunner",
        )

        with self.assertRaises(ValueError):
            config.validate()

    def test_executor_raises_when_beam_missing(self) -> None:
        config = self._config()
        executor = BeamPipelineExecutor()

        with mock.patch(
            "stscrapers.transformer.beam_pipeline.import_module",
            side_effect=ModuleNotFoundError("apache_beam"),
        ):
            transformer = BeamDataTransformer(
                config=config,
                transform_fn=lambda pcoll: pcoll,
                executor=executor,
            )
            with self.assertRaises(ImportError):
                transformer.process()


def sample_transform(values, suffix=""):
    return f"{values}{suffix}"


class TransformerCliTests(unittest.TestCase):
    def _args(self, **overrides) -> SimpleNamespace:
        defaults = {
            "config": None,
            "source_path": "input.csv",
            "source_format": "csv",
            "source_skip_header_lines": None,
            "source_compression": None,
            "source_remote": None,
            "output_file_path": "output.csv",
            "output_file_format": "csv",
            "output_file_shard_template": "",
            "output_file_suffix": None,
            "output_file_compression": None,
            "output_file_remote": None,
            "output_bigquery_dataset": None,
            "output_bigquery_table": None,
            "output_bigquery_schema": None,
            "output_bigquery_project": None,
            "runner": None,
            "project": None,
            "region": None,
            "temp_location": None,
            "staging_location": None,
            "streaming": None,
            "transform": "tests.test_transformer:sample_transform",
            "param": [],
        }
        defaults.update(overrides)
        return SimpleNamespace(**defaults)

    def test_build_config_from_args_without_config(self) -> None:
        args = self._args()
        config, config_data = build_config_from_args(args)
        self.assertEqual(config.source.path, "input.csv")
        self.assertEqual(config.output_file.path, "output.csv")
        self.assertEqual(config.runner, "DirectRunner")
        self.assertIsNone(config_data)

    def test_build_config_from_file_with_override(self) -> None:
        config_payload = {
            "source": {"path": "cfg_input.csv"},
            "output_file": {"path": "cfg_output.csv"},
            "runner": "DataflowRunner",
            "project": "cfg-project",
            "region": "us-central1",
            "temp_location": "gs://cfg-temp",
        }
        with tempfile.NamedTemporaryFile("w", delete=False) as handle:
            json.dump(config_payload, handle)
            config_path = handle.name
        self.addCleanup(lambda: os.unlink(config_path))
        args = self._args(
            config=config_path,
            source_path=None,
            output_file_path=None,
            project="override-project",
        )
        config, config_data = build_config_from_args(args)
        self.assertEqual(config.source.path, "cfg_input.csv")
        self.assertEqual(config.output_file.path, "cfg_output.csv")
        self.assertEqual(config.runner, "DataflowRunner")
        self.assertEqual(config.project, "override-project")
        self.assertEqual(config_data["source"]["path"], "cfg_input.csv")

    def test_parse_param_overrides_casts_values(self) -> None:
        params = parse_param_overrides(["threshold=0.5", "enabled=true", "label=beta"])
        self.assertEqual(params["threshold"], 0.5)
        self.assertTrue(params["enabled"])
        self.assertEqual(params["label"], "beta")

    def test_load_transform_callable_with_params(self) -> None:
        fn = load_transform_callable(
            "tests.test_transformer:sample_transform", {"suffix": "-ok"}
        )
        self.assertEqual(fn("data"), "data-ok")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import argparse
import json
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, Optional, Sequence, Type

from stscrapers.scrapers.logs.metadata import (
    DatasetStatusLogger,
    LocalCsvMetadataLogger,
    LocalDatasetStatusLogger,
    MetadataLogger,
)
from stscrapers.scrapers.base import DatasetScraper
from stscrapers.scrapers.models import (
    DatasetCheckResult,
    DatasetConfig,
    DatasetFetchResult,
    DatasetUpdateResult,
)
from stscrapers.scrapers.socrata import SocrataDatasetScraper
from stscrapers.storage.backends import LocalFileStorageBackend, StorageBackend


SCRAPER_REGISTRY: Dict[str, Type[DatasetScraper]] = {
    "socrata": SocrataDatasetScraper,
}


class ScraperManager:
    def __init__(
        self,
        storage_backend: Optional[StorageBackend] = None,
        metadata_logger: Optional[MetadataLogger] = None,
        status_logger: Optional[DatasetStatusLogger] = None,
    ) -> None:
        self.storage = storage_backend or LocalFileStorageBackend()
        self.metadata_logger = metadata_logger or LocalCsvMetadataLogger()
        self.status_logger = status_logger or LocalDatasetStatusLogger()

    def check_dataset(
        self, scraper: DatasetScraper, dataset: DatasetConfig
    ) -> DatasetCheckResult:
        result = scraper.check_updates(dataset)
        self.metadata_logger.record_check(dataset, result)
        self.status_logger.record_check(dataset, result)
        return result

    def update_dataset(
        self, scraper: DatasetScraper, dataset: DatasetConfig
    ) -> DatasetUpdateResult:
        fetch_result = scraper.fetch_dataset(dataset)
        stored_files = self._persist_files(dataset, fetch_result)
        metadata_payload = {
            "dataset_id": dataset.identifier,
            "downloaded_at": datetime.now(timezone.utc).isoformat(),
            "files": stored_files,
            "domain": dataset.domain,
            "extra": dataset.extra,
            "scraper_metadata": fetch_result.metadata,
        }
        metadata_path = self.storage.write_metadata(dataset, metadata_payload)
        update_result = DatasetUpdateResult(
            dataset_id=dataset.identifier,
            files=stored_files,
            metadata_path=metadata_path,
        )
        self.metadata_logger.record_update(dataset, update_result)
        self.status_logger.record_update(dataset, update_result)
        return update_result

    def _persist_files(
        self, dataset: DatasetConfig, fetch_result: DatasetFetchResult
    ) -> Sequence[Dict[str, object]]:
        stored = []
        for resource in fetch_result.files:
            file_path = self.storage.write_dataset_file(
                dataset, resource.format, resource.content
            )
            stored.append(
                {
                    "format": resource.format,
                    "path": str(file_path),
                    "bytes": len(resource.content),
                    "source_url": resource.source_url,
                }
            )
        return stored


def build_arg_parser(require_subcommand: bool = True) -> argparse.ArgumentParser:
    def __add_common_dataset_args(subparser: argparse.ArgumentParser) -> None:
        subparser.add_argument("--identifier", required=True, help="Dataset identifier (4x4)")
        subparser.add_argument("--domain", required=True, help="Dataset domain/host")
        subparser.add_argument(
            "--local-path",
            required=True,
            help="Local directory for storing downloaded assets",
        )

    parser = argparse.ArgumentParser(description="Scraper manager CLI")
    parser.add_argument(
        "--scraper",
        required=True,
        choices=SCRAPER_REGISTRY.keys(),
        help="Which scraper to run (e.g. socrata)",
    )
    parser.add_argument("--app-token", help="Optional scraper-specific token")
    subparsers = parser.add_subparsers(dest="command", required=require_subcommand)

    check_parser = subparsers.add_parser("check", help="Check dataset for updates")
    __add_common_dataset_args(check_parser)
    check_parser.add_argument(
        "--last-update",
        help="ISO timestamp for last successful update (e.g. 2024-01-01T00:00:00Z)",
    )

    update_parser = subparsers.add_parser("update", help="Download/export dataset files")
    __add_common_dataset_args(update_parser)
    update_parser.add_argument(
        "--download-formats",
        help="Comma separated list of download formats (default depends on scraper)",
    )

    return parser





def _build_dataset_from_args(args: argparse.Namespace) -> DatasetConfig:
    last_update_value = getattr(args, "last_update", None)
    last_update = (
        datetime.fromisoformat(last_update_value.replace("Z", "+00:00"))
        if last_update_value
        else None
    )
    download_formats_value = getattr(args, "download_formats", None)
    download_formats = (
        tuple(fmt.strip() for fmt in download_formats_value.split(","))
        if download_formats_value
        else None
    )
    return DatasetConfig(
        identifier=args.identifier,
        domain=args.domain,
        local_path=Path(args.local_path),
        last_local_update=last_update,
        download_formats=download_formats,
    )


def _create_scraper(args: argparse.Namespace) -> DatasetScraper:
    scraper_cls = SCRAPER_REGISTRY[args.scraper]
    if scraper_cls is SocrataDatasetScraper:
        return scraper_cls(app_token=args.app_token)
    return scraper_cls()  # type: ignore[return-value]


def main(argv: Optional[Sequence[str]] = None) -> int:

    def __serialize_check_result(result: DatasetCheckResult) -> Dict[str, object]:
        payload: Dict[str, object] = {
            "dataset_id": result.dataset_id,
            "has_updates": result.has_updates,
            "metadata": result.metadata,
        }
        payload["remote_updated_at"] = (
            result.remote_updated_at.isoformat() if result.remote_updated_at else None
        )
        return payload


    def __serialize_update_result(result: DatasetUpdateResult) -> Dict[str, object]:
        return {
            "dataset_id": result.dataset_id,
            "files": list(result.files),
            "metadata_path": str(result.metadata_path),
        }

    parser = build_arg_parser()
    args = parser.parse_args(argv)
    dataset = _build_dataset_from_args(args)
    scraper = _create_scraper(args)
    manager = ScraperManager()

    if args.command == "check":
        result = manager.check_dataset(scraper, dataset)
        print(json.dumps(__serialize_check_result(result), indent=2))
        return 0
    if args.command == "update":
        result = manager.update_dataset(scraper, dataset)
        print(json.dumps(__serialize_update_result(result), indent=2))
        return 0
    parser.error("Unknown command")
    return 1





if __name__ == "__main__":  # pragma: no cover
    raise SystemExit(main())



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class FileSourceConfig:
    """Represents an input file stored either locally or in cloud storage."""

    path: str
    file_format: str = "csv"
    skip_header_lines: int = 0
    compression: Optional[str] = None
    is_remote: bool = False

    def resolved_path(self) -> str:
        if self.is_remote:
            return self.path
        return str(Path(self.path).expanduser().resolve())

#TODO: Create SourceConfig class for retrieving file from Google Cloud Bucket

@dataclass
class FileSinkConfig:
    """Writes the transformed output to a file (local or remote)."""

    path: str
    file_format: str = "csv"
    shard_name_template: str = ""
    file_name_suffix: Optional[str] = None
    compression: Optional[str] = None
    is_remote: bool = False

    def resolved_path(self) -> str:
        if self.is_remote:
            return self.path
        return str(Path(self.path).expanduser().resolve())


@dataclass
class BigQuerySinkConfig:
    """Writes the transformed output to BigQuery."""

    dataset: str
    table: str
    schema: str
    project: Optional[str] = None
    create_disposition: str = "CREATE_IF_NEEDED"
    write_disposition: str = "WRITE_TRUNCATE"

    def table_spec(self) -> str:
        if self.project:
            return f"{self.project}:{self.dataset}.{self.table}"
        return f"{self.dataset}.{self.table}"


@dataclass
class TransformerConfig:
    """Holds runner and IO settings for a Beam-based transformation."""

    source: FileSourceConfig
    runner: str = "DirectRunner"
    output_file: Optional[FileSinkConfig] = None
    output_bigquery: Optional[BigQuerySinkConfig] = None
    project: Optional[str] = None
    region: Optional[str] = None
    temp_location: Optional[str] = None
    staging_location: Optional[str] = None
    streaming: bool = False

    def validate(self) -> None:
        if not (self.output_file or self.output_bigquery):
            raise ValueError("At least one output (file or BigQuery) must be configured.")
        if self.runner == "DataflowRunner":
            missing = [
                field_name
                for field_name, value in {
                    "project": self.project,
                    "temp_location": self.temp_location,
                    "region": self.region,
                }.items()
                if not value
            ]
            if missing:
                raise ValueError(
                    "DataflowRunner requires the following config fields: "
                    + ", ".join(missing)
                )
